import argparse
import http.server
import os
import socketserver
import sys
from pathlib import Path

from yabi.blog import Blog
from yabi.post import Post

DEFAULT_TEST_PORT = 9090
DEFAULT_TEST_HOST = 'localhost'


def parse_cli_arguments():
    parser = argparse.ArgumentParser(prog='yabi', description='A static blog site generator')

    subparsers = parser.add_subparsers(title='subcommands', dest='command', description='valid subcommands', required=True)

    parser_init = subparsers.add_parser('init', help='Creates a new yabi blog website')
    parser_init.add_argument('path', help='Initializes all the relevant files for the website on the input path')

    parser_build = subparsers.add_parser('build', help='Builds the website')
    parser_build.add_argument('--force', help='Force a clean rebuild of the entire website', action='store_true')

    subparsers.add_parser('test', help='Creates a local server to check the blog locally')

    return parser.parse_args()


def init(path: Path):
    blog = Blog(path)
    blog.create()
    print(f'New yabi blog successfully created on {path.resolve()}!')


def build(blog: Blog, force: bool):
    blog.load_config()
    all_public_posts = []
    needs_rebuild = False

    if len(list(blog.markdown_post_paths())) == 0:
        print(f'Error: No markdown posts found under the path {blog.posts_path}')
        return 1

    blog.create_base_website()

    if blog.is_config_file_updated() or force:
        print(f'The config.json file has been modified. Rebuilding whole site...')
        force = True

    for path in blog.markdown_post_paths():
        target_path = blog.get_post_target_html_path(path)
        post = Post(path, target_path)
        all_public_posts.append(post)
        if post.is_public() and (post.is_dirty(target_path) or force):
            print(f'Building post {post.source_path}...')
            blog.build_post(post)
            if not needs_rebuild:
                needs_rebuild = True

    # Cleanup: If a post was deleted after it had been published, then we need to delete the corresponding html file.
    for target_path in blog.orphan_target_paths():
        print(f'Deleting orphan page: {target_path}')
        target_path.unlink()

    if force or needs_rebuild:
        all_public_posts.sort(key=lambda x: x.date, reverse=True)
        print(f'Building index...')
        blog.build_home_page(all_public_posts)
        print(f'Building tag pages...')
        blog.build_tag_page(all_public_posts)
        print(f'Building archive pages...')
        blog.build_archive_page(all_public_posts)
        print(f'Done!')
    else:
        print('No new posts found!')

    if force or needs_rebuild:
        blog.update_last_build_file()


def serve(filepath_to_serve: Path):
    os.chdir(filepath_to_serve)
    with socketserver.TCPServer((DEFAULT_TEST_HOST, DEFAULT_TEST_PORT), http.server.SimpleHTTPRequestHandler) as httpd:
        try:
            print(f'Test server running on: http://{DEFAULT_TEST_HOST}:{DEFAULT_TEST_PORT}')
            httpd.serve_forever()
        except KeyboardInterrupt:
            print('Shutting Down Server')
            httpd.shutdown()
            httpd.server_close()


def execute():
    args = parse_cli_arguments()

    if args.command == 'init':
        init(Path(args.path))

    else:
        blog = Blog(Path('.'))
        if not blog.is_blog():
            print(f'Error: The current path "{blog.main_path}" does not contain a yabi blog')
            if not blog.posts_path.exists():
                print(f'Error: No "{blog.POSTS_DIR_NAME}" path found on {blog.main_path}')
            if not blog.data_path.exists():
                print(f'Error: No "{blog.DATA_DIR_NAME}" path found on {blog.main_path}')
            if not blog.config_path.exists():
                print(f'Error: No {blog.CONFIG_FILE_NAME} path found on {blog.main_path}')
            return 1

        if args.command == 'build':
            build(blog, args.force)

        elif args.command == 'test':
            serve(blog.website_path)


if __name__ == '__main__':
    sys.exit(execute())



import calendar
import datetime as dt
import json
import shutil
import sys
from collections.abc import Iterator
from importlib import resources
from pathlib import Path

from jinja2 import Environment, PackageLoader

from yabi.post import Post


class Blog:
    TEMPLATE_DIR_NAME = 'templates'
    STYLE_SHEET_DIR_NAME = 'style_sheets'
    WEBSITE_DIR_NAME = 'public'
    POSTS_DIR_NAME = 'posts'
    TAGS_DIR_NAME = 'tags'
    ARCHIVE_DIR_NAME = 'archive'
    DATA_DIR_NAME = 'data'
    POST_TEMPLATE = 'post.html'
    TAG_TEMPLATE = 'tag.html'
    ARCHIVE_TEMPLATE = 'archive.html'
    ALL_TAGS_TEMPLATE = 'all_tags.html'
    ALL_ARCHIVE_TEMPLATE = 'all_archive.html'
    INDEX_TEMPLATE = 'index.html'
    CSS_FILE_NAME = 'style.css'
    CONFIG_FILE_NAME = 'config.json'
    LAST_BUILD_FILE_NAME = '.yabi_last_build'
    HOME_MAX_POSTS = 10

    def __init__(self, main_path: Path):
        self.main_path = main_path.resolve().expanduser()

        self.website_path = main_path / self.WEBSITE_DIR_NAME
        self.website_posts_path = self.website_path / self.POSTS_DIR_NAME
        self.website_tags_path = self.website_path / self.TAGS_DIR_NAME
        self.website_archive_path = self.website_path / self.ARCHIVE_DIR_NAME

        self.posts_path = main_path / self.POSTS_DIR_NAME
        self.data_path = main_path / self.DATA_DIR_NAME
        self.templates_path = self.data_path / self.TEMPLATE_DIR_NAME
        self.style_sheets_path = self.data_path / self.STYLE_SHEET_DIR_NAME
        self.default_css_file_path = self.style_sheets_path / self.CSS_FILE_NAME
        self.last_build_file_path = self.main_path / self.LAST_BUILD_FILE_NAME

        self.template_environment = Environment(loader=PackageLoader('yabi'), trim_blocks=True, lstrip_blocks=True)
        self.template_environment.globals.update({'current_year': f'{dt.date.today().year}',
                                                  'website_path': self.website_path})
        self.config_path = main_path / self.CONFIG_FILE_NAME

    def create(self):
        if self.is_blog():
            print(f'Error! Input path {self.main_path.resolve()} seems to contain another yabi blog')
            sys.exit(1)
        elif not self.is_blog() and self.main_path.exists():
            print(f'Error! Input path {self.main_path.resolve()} already exists. Please choose a another path to create a yabi blog')
            sys.exit(1)

        self.main_path.mkdir(parents=True)
        self.posts_path.mkdir()
        self.data_path.mkdir()
        self.update_last_build_file()
        self._save_default_config()

    def load_config(self):
        """ Loads the config file and applies the globals to the environment """
        with self.config_path.open() as file:
            json_encoded = file.read()
        config = json.loads(json_encoded)
        self.template_environment.globals.update(config)

    def create_base_website(self):
        self.website_path.mkdir(exist_ok=True)
        self.website_posts_path.mkdir(exist_ok=True)
        self.website_tags_path.mkdir(exist_ok=True)
        self.website_archive_path.mkdir(exist_ok=True)
        shutil.copy(self.default_css_file_path, self.website_path)

    def update_last_build_file(self):
        """ Updates the modification time of the "last build" file which keeps track of the last build time"""
        self.last_build_file_path.touch(exist_ok=True)

    def is_config_file_updated(self) -> bool:
        if not self.last_build_file_path.exists():
            self.update_last_build_file()
            return True
        else:
            return self.config_path.stat().st_mtime > self.last_build_file_path.stat().st_mtime

    def is_blog(self) -> bool:
        """ Checks whether the current directory is a yabi blog, i.e., it has the relevant paths"""
        if self.posts_path.exists() and self.data_path.exists() and self.config_path.exists():
            return True
        else:
            return False

    def build_home_page(self, all_posts: list[Post]):
        index_template = self.template_environment.get_template(self.INDEX_TEMPLATE)
        pagination_base_path = self.website_path / 'index'
        for actual_index, previous_page, next_page, target_path, page_posts in self._iter_posts_pagination(all_posts, pagination_base_path):
            index_html = index_template.render(latest_posts=page_posts, index=actual_index,
                                               previous_page=previous_page, next_page=next_page)
            target_path.write_text(index_html)

    def build_tag_page(self, all_posts: list[Post]):
        all_tags = set([tag for post in all_posts for tag in post.tags])
        grouped_posts = [(tag, [post for post in all_posts if tag in post.tags]) for tag in all_tags]

        tag_template = self.template_environment.get_template(self.TAG_TEMPLATE)
        for tag, group in grouped_posts:
            pagination_base_path = self.website_tags_path / f'{tag}'
            for actual_index, previous_page, next_page, target_path, page_posts in self._iter_posts_pagination(group,
                                                                                                               pagination_base_path):
                tag_html = tag_template.render(tag=tag, latest_posts=page_posts, index=actual_index,
                                               previous_page=previous_page, next_page=next_page)
                target_path.write_text(tag_html)

        # Get font size increase depending on the amount of posts
        all_tags_with_sizes = [(tag, 100 + 0.5 * (len(posts))) for tag, posts in grouped_posts]

        all_tags_template = self.template_environment.get_template(self.ALL_TAGS_TEMPLATE)
        all_tags_html = all_tags_template.render(all_tags=all_tags_with_sizes)
        target_path = self.website_path / f'tags.html'
        target_path.write_text(all_tags_html)

    def build_archive_page(self, all_posts: list[Post]):
        all_months = list(set([(post.date.year, post.date.month) for post in all_posts]))
        all_months.sort(key=lambda x: x[0] * 10 - x[1], reverse=True)
        grouped_posts = {year: [] for year, _ in all_months}
        for year, month in all_months:
            month_tuple = (calendar.month_name[month],
                           [post for post in all_posts if (year == post.date.year and month == post.date.month)])
            grouped_posts[year].append(month_tuple)

        archive_template = self.template_environment.get_template(self.ARCHIVE_TEMPLATE)
        for year in grouped_posts:
            year_path = self.website_archive_path / f'{year}'
            year_path.mkdir(exist_ok=True)
            for month, group in grouped_posts[year]:
                pagination_base_path = year_path / f'{month.lower()}'
                for actual_index, previous_page, next_page, target_path, page_posts in self._iter_posts_pagination(group,
                                                                                                                   pagination_base_path):
                    tag_html = archive_template.render(month=month, year=year, latest_posts=page_posts, index=actual_index,
                                                       previous_page=previous_page, next_page=next_page)
                    target_path.write_text(tag_html)

        all_archive_template = self.template_environment.get_template(self.ALL_ARCHIVE_TEMPLATE)
        archive_html = all_archive_template.render(grouped_posts=grouped_posts)
        target_path = self.website_path / f'archive.html'
        target_path.write_text(archive_html)

    def markdown_post_paths(self) -> Iterator[Path]:
        return self.posts_path.rglob('*md')

    def orphan_target_paths(self) -> Iterator[Path]:
        """ Returns the html paths of the current build that do not have a corresponding markdown path """
        for target_path in self.website_posts_path.rglob('*.html'):
            expected_post_path = self.main_path / target_path.relative_to(self.website_path).with_suffix('.md')
            if not expected_post_path.exists():
                yield target_path

    def build_post(self, post: Post):
        post_template = self.template_environment.get_template(self.POST_TEMPLATE)
        html_content = post.get_content_in_html()
        html_page = post_template.render(post=post, content=html_content)
        post.target_path.parent.mkdir(exist_ok=True)
        post.target_path.write_text(html_page)

    def get_post_target_html_path(self, post_path: Path) -> Path:
        """ Target paths are named with the same name of the input markdown file name """
        return self.website_posts_path / post_path.parent.relative_to(self.posts_path) / f'{post_path.stem}.html'

    def _iter_posts_pagination(self, all_posts: list[Post], pagination_base_path: Path):
        if len(all_posts) <= self.HOME_MAX_POSTS:  # special case when the list of post is small enough
            actual_index = previous_page = next_page = None
            target_path = pagination_base_path.with_suffix('.html')
            yield actual_index, previous_page, next_page, target_path, all_posts
            return
        for idx, pos in enumerate(range(0, len(all_posts), self.HOME_MAX_POSTS)):
            page_posts = all_posts[pos:pos + self.HOME_MAX_POSTS]
            actual_index = idx + 1
            if idx == 0:
                target_path = pagination_base_path.with_suffix('.html')
                if len(all_posts) < self.HOME_MAX_POSTS:
                    actual_index = None
                    previous_page = None
                    next_page = None
                else:
                    pagination_base_path.mkdir(exist_ok=True)
                    previous_page = None
                    next_page = pagination_base_path / f'page_{actual_index + 1}.html'
            else:
                if idx == 1:
                    previous_page = pagination_base_path.with_suffix('.html')
                else:
                    previous_page = pagination_base_path / f'page_{actual_index - 1}.html'
                target_path = pagination_base_path / f'page_{actual_index}.html'
                # TODO: if the number of post is a multiple of HOME_MAX_POSTS then the below condition will be never satisfied
                if len(page_posts) < self.HOME_MAX_POSTS:
                    next_page = None
                else:
                    next_page = pagination_base_path / f'page_{actual_index + 1}.html'

            yield actual_index, previous_page, next_page, target_path, page_posts

    def _save_default_config(self):
        with resources.as_file(resources.files('yabi') / self.DATA_DIR_NAME) as data_directory:
            shutil.copytree(data_directory, self.data_path, dirs_exist_ok=True)

        for file in self.data_path.rglob('*'):  # refresh modification of all files to the current time
            file.touch()

        # TODO: think of adding a enum or something else for better control of all config variables
        config = {'website_name': self.main_path.resolve().name,
                  'website_author': '',
                  'website_description': '',
                  'website_keywords': '',
                  'website_root': '/'
                  }
        json_encoded = json.dumps(config)
        self.config_path.write_text(json_encoded)



"""
There are two requirements that markdown files should fulfill in order to be considered blog posts

    1. The first non-empty lines of the file should contain the metadata of the post: i.e., publish date, is it a draft,
       the tags, etc. The draft metadata is the only mandatory field.
    2. After the metadata, the next non-empty line should be a level 1 header, i.e., a title prepended by two "#" signs
"""
import datetime as dt
import re
from pathlib import Path

import markdown


class Post:
    MANDATORY_LABELS = ['draft']
    DEFAULT_TAG = 'blog'
    INVALID_LABELS = ['_metadata', 'target_path', 'source_path', 'title']

    TITLE_REGEXP = re.compile(r'^\s*#(?!#)\s*(.*?)\s*$', flags=re.MULTILINE)
    METADATA_REGEXP = re.compile(r'^\s*(\w+)\s*:\s*(.+?)\s*$', flags=re.MULTILINE)

    def __init__(self, source_path: Path, target_path: Path):
        self.source_path = source_path
        self.target_path = target_path
        self._metadata = self.parse_metadata()

    def is_dirty(self, target_path: Path) -> bool:
        """ Checks whether the post needs to be rebuilt """
        file_mtime = self.source_path.stat().st_mtime
        target_mtime = target_path.stat().st_mtime if target_path.exists() else 0
        return file_mtime > target_mtime

    def is_public(self) -> bool:
        return self._metadata['draft'] != 'yes'

    def __getattr__(self, item):
        return self._metadata[item]

    def get_content_in_html(self) -> str:
        """ Transforms content from markdown to html """
        with self.source_path.open() as file:
            raw_text = file.read()
        title = self._metadata['title']
        index = raw_text.find(title)
        if index == -1:
            raise ValueError(f'The title {title} was not found in the text')
        markdown_text = raw_text[index + len(title):].strip()
        return markdown.markdown(markdown_text)

    def parse_metadata(self) -> dict[str, str]:
        """
        Gets all the labels like "label: value" at the beginning of the post and also retrieve the title following
        this label
        """
        metadata = {}
        with self.source_path.open() as file:
            for raw_line in file:
                line = raw_line.strip()
                if line:
                    if match := self.METADATA_REGEXP.match(line):
                        key, value = match.group(1).lower(), match.group(2).lower()
                        if key in Post.INVALID_LABELS:
                            print(f'Invalid metadata label entry: "{key}". Ignoring...')
                            continue
                        else:
                            metadata.update({key: value})
                    elif match := self.TITLE_REGEXP.match(line):
                        metadata.update({'title': match.group(1)})
                        break
                    else:
                        break

        # Add default tag if nothing is found
        if 'tags' not in metadata:
            metadata.update({'tags': [Post.DEFAULT_TAG]})

        # Add default date if not found and prepend it to the file
        if 'date' not in metadata:
            today = dt.date.today()
            metadata.update({'date': today})
            with self.source_path.open() as file:
                content = file.read()
            with self.source_path.open(mode='w') as file:
                file.write(f'date: {today.isoformat()}\n')
                file.write(content)
        else:
            metadata.update({'date': dt.date.fromisoformat(metadata['date'])})

        # Check for errors
        missing_mandatory_labels = set(Post.MANDATORY_LABELS).difference(set(metadata))
        if missing_mandatory_labels:
            raise ValueError(f'The following mandatory label(s) is missing for the file {self.source_path}: {missing_mandatory_labels}')
        elif 'title' not in metadata:
            raise ValueError(f'No title found after the data labels for the file {self.source_path}')

        # Convert into the correct type the value of the keys
        for key, value in metadata.items():
            if isinstance(value, str) and value.startswith('[') and value.endswith(']'):
                actual_value = [list_element.strip() for list_element in metadata[key].strip(' []').split(',')]
            else:
                actual_value = value

            if key == 'tags' and isinstance(actual_value, str):
                actual_value = [actual_value]

            metadata.update({key: actual_value})

        return metadata

    def __eq__(self, other: 'Post'):
        if self.source_path == other.source_path and self.target_path == other.target_path and self._metadata == other._metadata:
            return True
        else:
            return False
